import os
from os.path import join as ospj
import time
import datetime
from munch import Munch
import logging

import numpy as np
import torch
import math
import cv2
import csv
from torch.utils import data
import torch.nn.functional as F
import torch.nn as nn
from PIL import Image

import util.utils as utils
from data.data_loader import InputFetcher
from data.data_loader import get_original_loader, get_val_loader
from model.build_models import build_model
from training.solver import Solver
from prune.Loss import DebiasedSupConLoss
from util.grad_cam import GradCAM
from torchvision.transforms import Resize, Compose, ToTensor, Normalize
import matplotlib.pyplot as plt
from util.cka import CKACalculator 
from copy import deepcopy


class PruneSolver(Solver):
    def __init__(self, args):
        super(PruneSolver, self).__init__(args)
        self.optims_main = Munch() # Used in retraining
        self.optims_mask = Munch() # Used in learning pruning parameter

        for net, m in self.nets.items():  
            prune_param = [p for n,p in m.named_parameters() if 'gumbel_pi' in n]
            main_param = [p for n,p in m.named_parameters() if 'gumbel_pi' not in n]

            if args.optimizer == 'Adam':
                self.optims_main[net] = torch.optim.Adam(
                    params=main_param,
                    lr=args.lr_main,
                    betas=(args.beta1, args.beta2),
                    weight_decay=0
                )
            elif args.optimizer == 'SGD':
                self.optims_main[net] = torch.optim.SGD(
                    main_param,
                    lr=args.lr_main,
                    momentum=0.9,
                    weight_decay=args.weight_decay
                )

            self.optims_mask[net] = torch.optim.Adam(
                prune_param,
                lr=args.lr_prune,
            )

        self.scheduler_main = Munch() # Used in retraining
        if not args.no_lr_scheduling:
            for net in self.nets.keys():
                self.scheduler_main[net] = torch.optim.lr_scheduler.StepLR(
                    self.optims_main[net], step_size=args.lr_decay_step_main, gamma=args.lr_gamma_main)

        self.con_criterion = DebiasedSupConLoss()

    def sparsity_regularizer(self, token='gumbel_pi'):
        reg = 0.
        for n, p in self.nets.classifier.named_parameters():
            if token in n:
                reg = reg + p.sum()
        return reg
    
    def sparsity_regularizer_a(self, token='gumbel_pi'):
        reg = 0.
        for n, p in self.nets.classifier_a.named_parameters():
            if token in n:
                reg = reg + p.sum()
        return reg
    
    def sparsity_regularizer_b(self, token='gumbel_pi'):
        reg = 0.
        for n, p in self.nets.classifier_b.named_parameters():
            if token in n:
                reg = reg + p.sum()
        return reg
    
    # IOU
    def sparsity_regularizer_ab(self, token='gumbel_pi'):
        reg = 0.
        a_activate_total = 0
        b_activate_total = 0
        ab_activate_total = 0
        Ratio_IoU_list = []
        A_ACT = []
        B_ACT = []
        AB_ACT = []
        for (n_a, p_a), (n_b, p_b) in zip(self.nets.classifier_a.named_parameters(), self.nets.classifier_b.named_parameters()):
            if n_a != n_b:
                print('models are inconsistent! ERROR!!!')
            if token in n_a:
                mask_a = torch.zeros_like(p_a)
                mask_b = torch.zeros_like(p_b)
                mask_a[p_a >= 0] = 1.
                a_activate = torch.sum(mask_a == 1).item()
                A_ACT.append(a_activate)

                mask_b[p_b >= 0] = 1.
                b_activate = torch.sum(mask_b == 1).item()
                B_ACT.append(b_activate)
                mask_ab = torch.logical_and(mask_a, mask_b).float()  
                ab_activate = torch.sum(mask_ab == 1).item()
                AB_ACT.append(ab_activate)

                a_activate_total += a_activate
                b_activate_total += b_activate
                ab_activate_total += ab_activate

                Ratio_IoU = ab_activate / (a_activate + b_activate - ab_activate + 1e-6)
                Ratio_IoU_list.append(Ratio_IoU)

                reg = reg + mask_ab.sum()

        Ratio_IoU_total = ab_activate_total / (a_activate_total + b_activate_total - ab_activate_total + 1e-6)
        Ratio_IoU_list.append(Ratio_IoU_total)

        return reg, Ratio_IoU_list, A_ACT, B_ACT,AB_ACT

    def save_wrong_idx_init(self, loader):
        self.nets.classifier.eval()
        self.nets.biased_classifier.eval()

        iterator = enumerate(loader)
        total_wrong, total_num = 0, 0
        wrong_idx = torch.empty(0).to(self.device)
        debias_idx = torch.empty(0).to(self.device)
        fname_full = []

        for _, (idx, data, attr, fname) in iterator:
            idx = idx.to(self.device)
            label = attr[:, 0].to(self.device)
            bias_label = attr[:, 1].to(self.device)
            data = data.to(self.device)

            with torch.no_grad():
                if self.args.select_with_GCE or self.args.data == 'celebA':
                    logit = self.nets.biased_classifier(data)
                else:
                    logit = self.nets.classifier(data)
                pred = logit.data.max(1, keepdim=True)[1].squeeze(1)
                wrong = (pred != label).long()

                debiased = (label != bias_label).long()

                total_wrong += wrong.sum()
                total_num += wrong.shape[0]
                wrong_idx = torch.cat((wrong_idx, idx[wrong == 1])).long()
                debias_idx = torch.cat((debias_idx, idx[debiased == 1])).long()

            fname_full = fname_full + list(fname) 

        assert total_wrong == len(wrong_idx)
        print('Number of wrong samples: ', total_wrong)
        self.confirm_pseudo_label_init(wrong_idx, debias_idx, total_num)
    
    def confirm_pseudo_label_init(self, wrong_idx, debias_idx, total_num):
        wrong_label = torch.zeros(total_num).to(self.device)
        debias_label = torch.zeros(total_num).to(self.device)

        for idx in wrong_idx:
            wrong_label[idx] = 1
        for idx in debias_idx:
            debias_label[idx] = 1

        total_wrong = (wrong_label != 0).sum().item()
        wrong_idx_path = ospj(self.args.checkpoint_dir, 'wrong_index.pth')

        if not self.args.supervised:
            torch.save(wrong_label, wrong_idx_path)
        else:
            torch.save(debias_label, wrong_idx_path)
        print('Saved wrong index label.')
        self.nets.classifier.train()  
        self.nets.biased_classifier.train()

    def update_wrong_idx(self, loader):
        self.nets.classifier_a.eval()
        self.nets.classifier_b.eval()

        iterator = enumerate(loader)   
        total_wrong, total_num = 0, 0
        wrong_idx = torch.empty(0).to(self.device)
        debias_idx = torch.empty(0).to(self.device)

        for _, (idx, data, attr, fname) in iterator:
            idx = idx.to(self.device)
            label = attr[:, 0].to(self.device) 
            bias_label = attr[:, 1].to(self.device) 
            data = data.to(self.device)

            with torch.no_grad():
                logit_a = self.nets.classifier_a(data)
                logit_b = self.nets.classifier_b(data)
                logit_ab = torch.stack([logit_a, logit_b], dim=0)

                weights = torch.tensor([0.1, 0.9]).to(self.device)
                weighted_pred = (logit_ab * weights.unsqueeze(1).unsqueeze(2)).sum(dim=0)
                final_pred = weighted_pred.argmax(dim=1)
                wrong = (final_pred != label).long()

                debiased = (label != bias_label).long() 

                total_wrong += wrong.sum()
                total_num += wrong.shape[0]
                wrong_idx = torch.cat((wrong_idx, idx[wrong == 1])).long()
                debias_idx = torch.cat((debias_idx, idx[debiased == 1])).long()

        assert total_wrong == len(wrong_idx)
        print('Number of wrong samples: ', total_wrong)
        self.confirm_update_label(wrong_idx, debias_idx, total_num)

    def confirm_update_label(self, wrong_idx, debias_idx, total_num):
        wrong_label = torch.zeros(total_num).to(self.device)
        debias_label = torch.zeros(total_num).to(self.device)

        for idx in wrong_idx:
            wrong_label[idx] = 1
        for idx in debias_idx:
            debias_label[idx] = 1

        wrong_idx_path = ospj(self.args.checkpoint_dir, 'wrong_index.pth')

        total_wrong = (wrong_label != 0).sum().item()
        Idea_num = int(total_wrong + total_wrong * 0.1)
        if total_wrong < Idea_num:
            num_to_select =  Idea_num - total_wrong
            not_wrong_idx = (wrong_label == 0).nonzero().squeeze()
            selected_idx = torch.randperm(len(not_wrong_idx))[:num_to_select]
            wrong_label[not_wrong_idx[selected_idx]] = 2

        if os.path.exists(wrong_idx_path):
            wrong_label_exist = torch.load(wrong_idx_path).to(self.device)
            wrong_label[(wrong_label_exist == 1)] = 1
            wrong_label[(wrong_label_exist == 0) & (wrong_label == 1)] = 2
            wrong_label[(wrong_label_exist == 2) & (wrong_label == 0)] = 0
            wrong_label[(wrong_label_exist == 2) & (wrong_label == 1)] = 1

        if not self.args.supervised:
            torch.save(wrong_label, wrong_idx_path)
        else:
            torch.save(debias_label, wrong_idx_path)
        print('Update wrong index label.')
        self.nets.classifier_a.train() 
        self.nets.classifier_b.train()

    def update_wrong_idx_cifar(self, bias_score_array, loader, iters, pseudo_every):
        self.nets.classifier_a.eval()
        self.nets.classifier_b.eval()

        iterator = enumerate(loader)   
        debias_idx = torch.empty(0).to(self.device)

        for _, (idx, data, attr, fname) in iterator:
            idx = idx.to(self.device)
            label = attr[:, 0].to(self.device)
            bias_label = attr[:, 1].to(self.device) 
            data = data.to(self.device)
            debiased = (label != bias_label).long()

            with torch.no_grad():
                logit_b = self.nets.classifier_b(data)
                bias_prob_b = nn.Softmax()(logit_b)[torch.arange(logit_b.size(0)), label]
                bias_score_b = 1 - bias_prob_b
                for i, v in enumerate(idx):
                    bias_score_array[v] += bias_score_b[i] * pseudo_every / iters  
                debias_idx = torch.cat((debias_idx, idx[debiased == 1])).long()
        self.nets.classifier_a.train()
        self.nets.classifier_b.train()
        return bias_score_array, debias_idx

    def confirm_update_label_cifar(self, bias_score_array, debias_idx, total_num):
        wrong_label = (bias_score_array > (self.args.tau + 0.05)).long()
        debias_label = torch.zeros(total_num).to(self.device)

        for idx in debias_idx:
            debias_label[idx] = 1

        wrong_idx_path = ospj(self.args.checkpoint_dir, 'wrong_index.pth')

        if os.path.exists(wrong_idx_path):
            wrong_label_exist = torch.load(wrong_idx_path).to(self.device)
            wrong_label[(wrong_label_exist == 1)] = 1
            wrong_label[(wrong_label_exist == 0) & (wrong_label == 1)] = 2
            wrong_label[(wrong_label_exist == 2) & (wrong_label == 0)] = 0
            wrong_label[(wrong_label_exist == 2) & (wrong_label == 1)] = 1

        total_wrong = (wrong_label != 0).sum().item()

        if not self.args.supervised:
            torch.save(wrong_label, wrong_idx_path)
        else:
            torch.save(debias_label, wrong_idx_path)
        print('Update wrong index label.')
        self.nets.classifier_a.train()
        self.nets.classifier_b.train()

    def train_PRUNE(self, iters):
        args = self.args
        nets = self.nets
        save_ratio = []
        save_A_ACT = []
        save_B_ACT = []
        save_AB_ACT = []

        total_num = len(self.loaders.trainset)
        bias_score_array = torch.zeros(total_num).to(self.device)
        pseudo_every = int(total_num / args.batch_size) 

        self._load_checkpoint(args.pretrain_iter, 'pretrain', which='classifier')
        self.nets.classifier_a.load_state_dict(self.nets.classifier.state_dict())
        self.nets.classifier_b.load_state_dict(self.nets.classifier.state_dict())

        optims = self.optims_mask

        wrong_label = torch.load(ospj(self.args.checkpoint_dir, 'wrong_index.pth')).to(self.device)

        if wrong_num_ones != 0:
            wrong_Ratio = wrong_num_ones / wrong_num_zeros
        else:
            wrong_Ratio = 1 / 2000

        upweight_a = torch.ones_like(wrong_label)
        upweight_a[wrong_label == 1] = int((self.num_classes - 1) / wrong_Ratio)
        upweight_a[wrong_label == 2] = 40
        sampling_weight_a = upweight_a if not args.uniform_weight else torch.ones_like(wrong_label)
        a_loader = get_original_loader(args, sampling_weight=sampling_weight_a)

        upweight_b = torch.ones_like(wrong_label)
        upweight_b[wrong_label == 1] = 0
        upweight_b[wrong_label == 2] = 0
        sampling_weight_b = upweight_b if not args.uniform_weight else torch.ones_like(wrong_label)
        b_loader = get_original_loader(args, sampling_weight=sampling_weight_b)       

        fetcher_a = InputFetcher(a_loader)
        fetcher_b = InputFetcher(b_loader)
        fetcher_val = self.loaders.val
        fetcher_train = self.loaders.train
        start_time = time.time()

        self.nets.classifier_a.pruning_switch(True)
        self.nets.classifier_b.pruning_switch(True)

        for i in range(iters):
            inputs_a = next(fetcher_a)
            idx_a, x_a, label_a, fname_a = inputs_a.index, inputs_a.x, inputs_a.y, inputs_a.fname
            bias_label_a = torch.index_select(wrong_label, 0, idx_a.long())

            pred_a, feature_a = self.nets.classifier_a(x_a, feature=True)
            loss_main_a = self.criterion(pred_a, label_a).mean()
            loss_reg_a = self.sparsity_regularizer_a()
            loss_con_a = self.con_criterion(F.normalize(feature_a, dim=1).unsqueeze(1), label_a, bias_label_a)
            loss_a = loss_main_a + args.lambda_sparse * loss_reg_a  + args.lambda_con_prune * loss_con_a 

            inputs_b = next(fetcher_b)
            idx_b, x_b, label_b, fname_b = inputs_b.index, inputs_b.x, inputs_b.y, inputs_b.fname
            bias_label_b = torch.index_select(wrong_label, 0, idx_b.long())

            pred_b, feature_b = self.nets.classifier_b(x_b, feature=True)
            if self.args.data != 'cifar10c' and self.args.data != 'bffhq':
                loss_main_b = self.bias_criterion(pred_b, label_b).mean()
            else:
                loss_main_b = self.bias_criterion(pred_b, label_b)  # bffhq
                # loss_main_b = self.criterion(pred_b, label_b)  # cifar
                bias_prob_b = nn.Softmax()(pred_b)[torch.arange(pred_b.size(0)), label_b]
                loss_main_b = loss_main_b[bias_prob_b > args.eta].mean()
            loss_reg_b = self.sparsity_regularizer_b()
            loss_b = loss_main_b + args.lambda_sparse * loss_reg_b 

            loss_mask, Ratio_IoU_list, A_ACT, B_ACT, AB_ACT = self.sparsity_regularizer_ab()
            loss_sum = loss_a + loss_b + args.lambda_sparse * loss_mask
            save_ratio.append(Ratio_IoU_list)
            save_A_ACT.append(A_ACT)
            save_B_ACT.append(B_ACT)
            save_AB_ACT.append(AB_ACT)

            self._reset_grad()
            optims.classifier_a.zero_grad()
            loss_a.backward(retain_graph=True)
            optims.classifier_b.zero_grad()
            loss_b.backward(retain_graph=True)
            loss_sum.backward()
            optims.classifier_a.step()
            optims.classifier_b.step()

            if (i+1) % args.print_every == 0:
                elapsed = time.time() - start_time
                elapsed = str(datetime.timedelta(seconds=elapsed))[:-7]
                log_a = "Elapsed time [%s], Iteration [%i/%i], LR [%.4f], "\
                    "Loss_main_a [%.6f] Loss_reg_a [%.6f] Loss_con_a [%.6f] Loss_a [%.6f] Ratio_IoU_list [%s] A_ACT [%s]" % (elapsed, i+1, iters,
                                                                          optims.classifier_a.param_groups[-1]['lr'],
                                                                          loss_main_a.item(),
                                                                          loss_reg_a.item(),
                                                                          loss_con_a.item(),
                                                                          loss_a.item(),
                                                                          str(Ratio_IoU_list),
                                                                          str(A_ACT))
                print(log_a)
                logging.info(log_a)
                log_b = "Elapsed time [%s], Iteration [%i/%i], LR [%.4f], "\
                    "Loss_main_b [%.6f] Loss_reg_b [%.6f] Loss_b [%.6f] AB_ACT [%s] B_ACT [%s] Loss_mask [%6f]" % (elapsed, i+1, iters,
                                                                          optims.classifier_b.param_groups[-1]['lr'],
                                                                          loss_main_b.item(),
                                                                          loss_reg_b.item(),
                                                                          loss_b.item(),
                                                                          str(AB_ACT),
                                                                          str(B_ACT),
                                                                          loss_mask.item())
                print(log_b)
                logging.info(log_b)

            if (i+1) % args.eval_every == 0:
                total_a = 0
                active_a = 0
                layerwise_a = {}
                for n_a, p_a in self.nets.classifier_a.named_parameters():
                    if 'gumbel_pi' in n_a:
                        active_n_a = (p_a>=0).sum().item()
                        total_n_a = torch.ones_like(p_a).sum().detach().item()
                        layerwise_a[n_a] = active_n_a / total_n_a
                        total_a += total_n_a
                        active_a += active_n_a
                        if active_n_a == 0: print('Warning: Dead layer')
                ratio_a = active_a / total_a
                logging.info('Ratio_a of active to total_a: %f' % ratio_a)
                self.valid_logger.append(ratio_a, which='ratio')
                self.valid_logger.append(layerwise_a, which='layerwise_ratio')

                total_b = 0
                active_b = 0
                layerwise_b = {}
                for n_b, p_b in self.nets.classifier_b.named_parameters():
                    if 'gumbel_pi' in n_b:
                        active_n_b = (p_b>=0).sum().item()
                        total_n_b = torch.ones_like(p_b).sum().detach().item()
                        layerwise_b[n_b] = active_n_b / total_n_b
                        total_b += total_n_b
                        active_b += active_n_b
                        if active_n_b == 0: print('Warning: Dead layer')
                ratio_b = active_b / total_b
                logging.info('Ratio_b of active to total_b: %f' % ratio_b)
                self.valid_logger.append(ratio_b, which='ratio')
                self.valid_logger.append(layerwise_b, which='layerwise_ratio')

                self.nets.classifier_a.pruning_switch(False)
                self.nets.classifier_a.freeze_switch(True)
                total_acc_a, valid_attrwise_acc_a = self.validation(fetcher_val, which='classifier_a')
                self.report_validation(valid_attrwise_acc_a, total_acc_a, i, which='prune_a Test set')
                total_acc_a, valid_attrwise_acc_a = self.validation(fetcher_train, which='classifier_a')
                self.report_validation(valid_attrwise_acc_a, total_acc_a, i, which='prune_a Train set') 
                self.nets.classifier_a.pruning_switch(True)
                self.nets.classifier_a.freeze_switch(False)

                self.nets.classifier_b.pruning_switch(False)
                self.nets.classifier_b.freeze_switch(True)
                total_acc_b, valid_attrwise_acc_b = self.validation(fetcher_val, which='classifier_b')
                self.report_validation(valid_attrwise_acc_b, total_acc_b, i, which='prune_b Test set')
                total_acc_b, valid_attrwise_acc_b = self.validation(fetcher_train, which='classifier_b')
                self.report_validation(valid_attrwise_acc_b, total_acc_b, i, which='prune_b Train set')
                self.nets.classifier_a.pruning_switch(True)
                self.nets.classifier_a.freeze_switch(False)
                self.nets.classifier_b.pruning_switch(True)
                self.nets.classifier_b.freeze_switch(False)

            if (i+1) % (pseudo_every * 3)  == 0 and (i+1) < iters: 
                self.nets.classifier_a.pruning_switch(False)
                self.nets.classifier_a.freeze_switch(True)
                self.nets.classifier_b.pruning_switch(False)
                self.nets.classifier_b.freeze_switch(True)
                main_param_values = {}
                for net, m in self.nets.items():
                    main_param = [p for n,p in m.named_parameters() if 'gumbel_pi' not in n]
                    main_param_values[net] = [p.clone().detach() for p in main_param]
                if self.args.data != 'cifar10c' and self.args.data != 'bffhq':
                    self.retrain(pseudo_every)
                    self.update_wrong_idx(fetcher_train)
                else:
                    self.retrain(pseudo_every * 2)
                for net, m in self.nets.items():
                    main_param = [p for n,p in m.named_parameters() if 'gumbel_pi' not in n]
                    for p, v in zip(main_param, main_param_values[net]):
                        p.data.copy_(v)
                self.nets.classifier_a.pruning_switch(True)
                self.nets.classifier_a.freeze_switch(False)
                self.nets.classifier_b.pruning_switch(True)
                self.nets.classifier_b.freeze_switch(False)

                wrong_label = torch.load(ospj(self.args.checkpoint_dir, 'wrong_index.pth')).to(self.device)
                if wrong_num_ones != 0:
                    wrong_Ratio = wrong_num_ones / wrong_num_zeros
                else:
                    wrong_Ratio = 1 / 2000

                upweight_a = torch.ones_like(wrong_label)
                upweight_a[wrong_label == 1] = int((self.num_classes - 1) / wrong_Ratio)
                upweight_a[wrong_label == 2] = 40
                sampling_weight_a = upweight_a if not args.uniform_weight else torch.ones_like(wrong_label)
                a_loader = get_original_loader(args, sampling_weight=sampling_weight_a)

                upweight_b = torch.ones_like(wrong_label)
                upweight_b[wrong_label == 1] = 0
                upweight_b[wrong_label == 2] = 0
                sampling_weight_b = upweight_b if not args.uniform_weight else torch.ones_like(wrong_label)
                b_loader = get_original_loader(args, sampling_weight=sampling_weight_b)

                fetcher_a = InputFetcher(a_loader)
                fetcher_b = InputFetcher(b_loader)

            if (i+1) % args.save_every == 0:
                self._save_checkpoint(step=i+1, token='prune')
        self._load_checkpoint(args.pruning_iter, 'prune')
        self.nets.classifier.pruning_switch(True)
        for (n, p), (n_a, p_a), (n_b, p_b) in zip(self.nets.classifier.named_parameters(), self.nets.classifier_a.named_parameters(), self.nets.classifier_b.named_parameters()):
            if n_a != n_b and n != n_b:
                print('models are inconsistent! ERROR!!!')
            if 'gumbel_pi' in n_a:
                mask_a = torch.zeros_like(p_a)
                mask_b = torch.zeros_like(p_b)
                mask_a[p_a >= 0] = 1.
                mask_b[p_b >= 0] = 1.
                mask_ab = torch.logical_and(mask_a, mask_b).float() 
                only_a = torch.logical_and(mask_a, torch.logical_not(mask_b)).float()

                p_ab = torch.zeros_like(p_a)
                p_ab = torch.where(mask_ab == 1, p_a, p_ab)
                p_ab = torch.where(only_a == 1, (p_a + p_b), p_ab)
                p_ab = torch.where(mask_a == 0, p_a, p_ab)
                p.data.copy_(p_ab)

        self.nets.classifier.pruning_switch(False)
        self.nets.classifier.freeze_switch(True)
        total_acc, valid_attrwise_acc = self.validation(fetcher_val, which='main')
        self.report_validation(valid_attrwise_acc, total_acc, i, which='prune')
        total = 0
        active = 0
        layerwise = {}
        for n, p in self.nets.classifier.named_parameters():
            if 'gumbel_pi' in n:
                active_n = (p>=0).sum().item()
                total_n = torch.ones_like(p).sum().detach().item()
                layerwise[n] = active_n / total_n

                log = "active_n [%.2f] layerwise [%.6f] " % (active_n,layerwise[n])
                print(log)
                logging.info(log)

                total += total_n
                active += active_n
                if active_n == 0: print('Warning: Dead layer')

        ratio = active / total
        print('ratio in classifier :', ratio)
        logging.info('Ratio of active to total in classifier: %f' % ratio)
        self._save_checkpoint(step=args.pruning_iter, token='prune')

        headers = ['Iteration', 'Ratio_IoU_list', 'A_ACT', 'B_ACT', 'AB_ACT']
        data = list(zip(range(len(save_ratio)), save_ratio, save_A_ACT, save_B_ACT, save_AB_ACT))
        with open(os.path.join(self.args.result_dir, 'ratio.csv'), 'w', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(headers)
            for row in data:
                writer.writerow(row)


    def retrain(self, iters, freeze=True):
        args = self.args
        nets = self.nets
        optims = self.optims_main # Train only weight parameter

        total_num = len(self.loaders.trainset)
        bias_score_array = torch.zeros(total_num).to(self.device) 
        pseudo_every = int(total_num / args.batch_size)

        wrong_label = torch.load(ospj(self.args.checkpoint_dir, 'wrong_index.pth')).to(self.device)

        if wrong_num_ones != 0:
            wrong_Ratio = wrong_num_ones / wrong_num_zeros
        else:
            wrong_Ratio = 1 / 2000

        upweight = torch.ones_like(wrong_label)
        upweight_loader = get_original_loader(args, sampling_weight=upweight)

        upweight_a = torch.ones_like(wrong_label)
        upweight_a[wrong_label == 1] = int((self.num_classes - 1) / wrong_Ratio)
        upweight_a[wrong_label == 2] = 40
        a_loader = get_original_loader(args, sampling_weight=upweight_a)

        upweight_b = torch.ones_like(wrong_label)
        upweight_b[wrong_label == 1] = 0
        upweight_b[wrong_label == 2] = 1
        b_loader = get_original_loader(args, sampling_weight=upweight_b) 
    
        fetcher = InputFetcher(upweight_loader)
        fetcher_a = InputFetcher(a_loader)
        fetcher_b = InputFetcher(b_loader)
        fetcher_val = self.loaders.val
        fetcher_train = self.loaders.train
        start_time = time.time()

        self.nets.classifier.pruning_switch(False)
        self.nets.classifier.freeze_switch(freeze)
        self.nets.classifier_a.pruning_switch(False)
        self.nets.classifier_a.freeze_switch(freeze)
        self.nets.classifier_b.pruning_switch(False)
        self.nets.classifier_b.freeze_switch(freeze)     


        for i in range(iters):
            inputs = next(fetcher)
            idx, x, label, fname = inputs.index, inputs.x, inputs.y, inputs.fname
            bias_label = torch.index_select(wrong_label, 0, idx.long())

            pred, feature = self.nets.classifier(x, feature=True)  
            loss_main = self.criterion(pred, label).mean() 
            loss_con = self.con_criterion(F.normalize(feature, dim=1).unsqueeze(1), label, bias_label)
            loss = loss_main + args.lambda_con_retrain * loss_con 

            inputs_a = next(fetcher_a)
            idx_a, x_a, label_a, fname_a = inputs_a.index, inputs_a.x, inputs_a.y, inputs_a.fname
            bias_label_a = torch.index_select(wrong_label, 0, idx_a.long())

            pred_a, feature_a = self.nets.classifier_a(x_a, feature=True)
            loss_main_a = self.criterion(pred_a, label_a).mean()
            loss_con_a = self.con_criterion(F.normalize(feature_a, dim=1).unsqueeze(1), label_a, bias_label_a)
            loss_a = loss_main_a + args.lambda_con_prune * loss_con_a 

            inputs_b = next(fetcher_b)
            idx_b, x_b, label_b, fname_b = inputs_b.index, inputs_b.x, inputs_b.y, inputs_b.fname
            bias_label_b = torch.index_select(wrong_label, 0, idx_b.long())

            pred_b, feature_b = self.nets.classifier_b(x_b, feature=True)
            if self.args.data != 'cifar10c' and self.args.data != 'bffhq':
                loss_main_b = self.bias_criterion(pred_b, label_b).mean()
            else:
                loss_main_b = self.bias_criterion(pred_b, label_b)
                # loss_main_b = self.criterion(pred_b, label_b)
                bias_prob_b = nn.Softmax()(pred_b)[torch.arange(pred_b.size(0)), label_b]
                loss_main_b = loss_main_b[bias_prob_b > args.eta].mean()
            loss_b = loss_main_b 

            self._reset_grad()
            loss.backward()
            loss_a.backward()
            loss_b.backward()
            optims.classifier.step()
            optims.classifier_a.step()
            optims.classifier_b.step()

            # print out log info
            if (i+1) % args.print_every == 0:
                elapsed = time.time() - start_time
                elapsed = str(datetime.timedelta(seconds=elapsed))[:-7]
                log = "Classifier Elapsed time [%s], Iteration [%i/%i], LR [%.4f], "\
                    "Loss_main [%.6f] Loss_con [%.6f] " % (elapsed, i+1, iters,
                                                           optims.classifier.param_groups[-1]['lr'],
                                                           loss_main.item(),
                                                           loss_con.item())
                # print(log)
                # logging.info(log)
                elapsed = time.time() - start_time
                elapsed = str(datetime.timedelta(seconds=elapsed))[:-7]
                log = "Classifier A Elapsed time [%s], Iteration [%i/%i], LR [%.4f], "\
                    "Loss_main_a [%.6f] Loss_con_a [%.6f] " % (elapsed, i+1, iters,
                                                           optims.classifier_a.param_groups[-1]['lr'],
                                                           loss_main_a.item(),
                                                           loss_con_a.item())
                print(log)
                # logging.info(log)
                elapsed = time.time() - start_time
                elapsed = str(datetime.timedelta(seconds=elapsed))[:-7]
                log = "Classifier B Elapsed time [%s], Iteration [%i/%i], LR [%.4f], "\
                    "Loss_main_b [%.6f] " % (elapsed, i+1, iters,
                                                           optims.classifier_b.param_groups[-1]['lr'],
                                                           loss_main_b.item())
                print(log)

            if (i+1) % args.save_every_retrain == 0:
                self._save_checkpoint(step=i+1, token='retrain')

            if (i+1) % args.eval_every_retrain == 0:
                total_acc, valid_attrwise_acc = self.validation(fetcher_val, which='main')
                self.report_validation(valid_attrwise_acc, total_acc, i, which='retrain')
                self.valid_logger.append(total_acc.item(), which='retrain')
                self.valid_logger.append(valid_attrwise_acc, which='groupwise_acc')

                total_acc_a, valid_attrwise_acc_a = self.validation(fetcher_val, which='classifier_a')
                self.report_validation(valid_attrwise_acc_a, total_acc_a, i, which='retrain Test set A')
                total_acc_b, valid_attrwise_acc_b = self.validation(fetcher_val, which='classifier_b')
                self.report_validation(valid_attrwise_acc_b, total_acc_b, i, which='retrain Test set B')

                total_acc_train_a, valid_attrwise_acc_train_a = self.validation(fetcher_train, which='classifier_a')
                self.report_validation(valid_attrwise_acc_train_a, total_acc_train_a, i, which='retrain Train set A')
                total_acc_train_b, valid_attrwise_acc_train_b = self.validation(fetcher_train, which='classifier_b')
                self.report_validation(valid_attrwise_acc_train_b, total_acc_train_b, i, which='retrain Train set B')

            if (i+1) % pseudo_every == 0 and (self.args.data == 'cifar10c' or self.args.data == 'bffhq'): 
                bias_score_array, debias_idx = self.update_wrong_idx_cifar(bias_score_array, fetcher_train, iters, pseudo_every)

            if not self.args.no_lr_scheduling:
                self.scheduler_main.classifier.step()
                self.scheduler_main.classifier_a.step()
                self.scheduler_main.classifier_b.step()

        if self.args.data == 'cifar10c' or self.args.data == 'bffhq':
            self.confirm_update_label_cifar(bias_score_array, debias_idx, total_num)

    def train(self):
        logging.info('=== Start training ===')

        args = self.args
        loader = self.loaders.train

        try:
            self._load_checkpoint(args.pretrain_iter, 'pretrain')
            print('Pretrained ckpt exists. Checking upweight index ckpt...')
        except:
            print('Start pretraining...')
            self.train_ERM(args.pretrain_iter)
            self._load_checkpoint(args.pretrain_iter, 'pretrain')

        if os.path.exists(ospj(args.checkpoint_dir, 'wrong_index.pth')):
            print('Upweight ckpt exists.')
        else:
            print('Upweight ckpt does not exist. Creating...')
            if args.pseudo_label_method == 'wrong' or args.mode == 'JTT':
                if args.earlystop_iter is not None: self._load_checkpoint(args.earlystop_iter, 'pretrain')
                self.save_wrong_idx_init(loader)
                self._load_checkpoint(args.pretrain_iter, 'pretrain')
            else:
                raise ValueError('No upweight ckpt')

        assert os.path.exists(ospj(args.checkpoint_dir, 'wrong_index.pth'))

        if args.mode != 'JTT':
            try:
                self._load_checkpoint(args.pruning_iter, 'prune')
                print('Pruning parameter ckpt exists. Start retraining...')
            except:
                print('Pruning parameter ckpt does not exist. Start pruning...')
                self.train_PRUNE(args.pruning_iter)

        self.valid_logger.save()

        if self.args.reinitialize:
            reinit_dict = torch.load(ospj(args.checkpoint_dir, '{:06d}_{}_nets.ckpt'.format(0, 'initial')))['classifier']
            mask_dict = torch.load(ospj(args.checkpoint_dir, '{:06d}_{}_nets.ckpt'.format(args.pruning_iter, 'prune')))['classifier']
            pruning_dict = {k: v for k, v in mask_dict.items() if 'gumbel_pi' in k}
            reinit_dict.update(pruning_dict) # Update pruning parameters only
            self.nets.classifier.load_state_dict(reinit_dict)
            print('Reinitialized model from ', ospj(args.checkpoint_dir, '{:06d}_{}_nets.ckpt'.format(0, 'initial')))

        if args.mode == 'JTT':
            self._load_checkpoint(0, 'initial')

        self.retrain(args.retrain_iter, freeze=True if args.mode != 'JTT' else False)
        self.valid_logger.save()
        print('Finished training')

    def evaluate(self):
        self._load_checkpoint(self.args.pretrain_iter, 'pretrain')
        self.nets.classifier.pruning_switch(False)
        self.nets.classifier.freeze_switch(True)
        self.nets.classifier_a.pruning_switch(False)
        self.nets.classifier_a.freeze_switch(True)
        self.nets.classifier_b.pruning_switch(False)
        self.nets.classifier_b.freeze_switch(True)
        model_m = deepcopy(self.nets.classifier)
        model_m.eval()

        fetcher_val = self.loaders.val
        self._load_checkpoint(self.args.retrain_iter, 'retrain')
        self.nets.classifier.pruning_switch(False)
        self.nets.classifier.freeze_switch(True)
        self.nets.classifier_a.pruning_switch(False)
        self.nets.classifier_a.freeze_switch(True)
        self.nets.classifier_b.pruning_switch(False)
        self.nets.classifier_b.freeze_switch(True)

        total_acc, valid_attrwise_acc = self.validation(fetcher_val,which='main')
        self.report_validation(valid_attrwise_acc, total_acc, 0, which='Main Test', save_in_result=True)
        total_acc, valid_attrwise_acc = self.validation(fetcher_val,which='classifier_a')
        self.report_validation(valid_attrwise_acc, total_acc, 0, which='Classifier A Test', save_in_result=True)
        total_acc, valid_attrwise_acc = self.validation(fetcher_val,which='classifier_b')
        self.report_validation(valid_attrwise_acc, total_acc, 0, which='Classifier B Test', save_in_result=True)

        self._tsne(fetcher_val)

        fetcher_cka = self.loaders.cka
        self.nets.classifier_a.eval()
        self.nets.classifier_b.eval()
        self.nets.classifier.eval()
        model_b = deepcopy(self.nets.classifier_b)
        model_b.eval()
        model_a = deepcopy(self.nets.classifier_a)
        model_a.eval()

        calculator = CKACalculator(model1=model_a, model2=model_b, dataloader=fetcher_cka)
        cka_output = calculator.calculate_cka_matrix()
        print(f"CKA output size: {cka_output.size()}")
        plt.rcParams['figure.figsize'] = (10, 10)
        plt.xlabel('Classifier A')
        plt.ylabel('Classifier B')
        # plt.xticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        # plt.yticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        plt.xticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        plt.yticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        ax = plt.gca()
        ax.xaxis.set_ticks_position('top')
        plt.imshow(cka_output.cpu().numpy(), cmap='inferno')
        save_cka_path = os.path.join(self.args.log_dir, 'cka_ab.pdf')
        plt.savefig(save_cka_path, bbox_inches='tight')
        
        calculator = CKACalculator(model1=model_m, model2=model_a, dataloader=fetcher_cka)
        cka_output = calculator.calculate_cka_matrix()
        print(f"CKA output size: {cka_output.size()}")
        plt.rcParams['figure.figsize'] = (10, 10)
        plt.xlabel('Main Classifier')
        plt.ylabel('Classifier A')
        # plt.xticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        # plt.yticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        plt.xticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        plt.yticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        ax = plt.gca()
        ax.xaxis.set_ticks_position('top')
        plt.imshow(cka_output.cpu().numpy(), cmap='inferno')
        save_cka_path = os.path.join(self.args.log_dir, 'cka_am.pdf')
        plt.savefig(save_cka_path, bbox_inches='tight')

        calculator = CKACalculator(model1=self.nets.classifier_b, model2=model_b, dataloader=fetcher_cka)
        cka_output = calculator.calculate_cka_matrix()
        print(f"CKA output size: {cka_output.size()}")
        plt.rcParams['figure.figsize'] = (10, 10)
        plt.xlabel('Classifier B')
        plt.ylabel('Classifier B')
        # plt.xticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        # plt.yticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        plt.xticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        plt.yticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        ax = plt.gca()
        ax.xaxis.set_ticks_position('top')
        plt.imshow(cka_output.cpu().numpy(), cmap='inferno')
        save_cka_path = os.path.join(self.args.log_dir, 'cka_bb.pdf')
        plt.savefig(save_cka_path, bbox_inches='tight')

        calculator = CKACalculator(model1=model_m, model2=model_b, dataloader=fetcher_cka)
        cka_output = calculator.calculate_cka_matrix()
        print(f"CKA output size: {cka_output.size()}")
        plt.rcParams['figure.figsize'] = (10, 10)
        plt.xlabel('Main Classifier')
        plt.ylabel('Classifier B')
        # plt.xticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        # plt.yticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        plt.xticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        plt.yticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        ax = plt.gca()
        ax.xaxis.set_ticks_position('top')
        plt.imshow(cka_output.cpu().numpy(), cmap='inferno')
        save_cka_path = os.path.join(self.args.log_dir, 'cka_bm.pdf')
        plt.savefig(save_cka_path)
        plt.savefig(save_cka_path, bbox_inches='tight')

        calculator = CKACalculator(model1=self.nets.classifier_a, model2=model_a, dataloader=fetcher_cka)
        cka_output = calculator.calculate_cka_matrix()
        print(f"CKA output size: {cka_output.size()}")
        plt.rcParams['figure.figsize'] = (10, 10)
        plt.xlabel('Classifier A')
        plt.ylabel('Classifier A')
        # plt.xticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        # plt.yticks(list(range(0,7)), labels= ['conv1','bn1', 'conv2', 'bn2', 'conv3', 'bn3', 'avgpool'])
        plt.xticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        plt.yticks([4, 14, 26, 38, 51], labels= ['layer1','layer2', 'layer3', 'layer4', 'fc'])
        ax = plt.gca()
        ax.xaxis.set_ticks_position('top')
        plt.imshow(cka_output.cpu().numpy(), cmap='inferno')
        save_cka_path = os.path.join(self.args.log_dir, 'cka_aa.pdf')
        plt.savefig(save_cka_path)
        plt.savefig(save_cka_path, bbox_inches='tight')
